#####################
# Event definitions #
#####################

aging <- ManipulationEvent$new(
    name = "Aging",
    description = "Normal aging of individuals",
    parameters = list(
        day = 1.0 / 365.0
    ),
    mechanism = expression({
        status[alive == 1L, age := age + day]
    })
)

other_mortality <- ManipulationEvent$new(
    name = "Other Mortality",
    description = "Risk of death when no stroke has occurred",
    parameters = list(
        shape = 13.16346, 
        scale = 165.34514
    ),
    mechanism = expression({
        u <-  status[alive == 1L & stroke == 0L, rbern(1 - pweibull(age, shape = shape, scale = scale))]
        status[alive == 1L & stroke == 0L, `:=` (alive = u, cause_of_death = 1L - u)]
    })
)

get_stroke <- ManipulationEvent$new(
    name = "Stroke",
    description = "Risk of suffering a stroke",
    parameters = list(
                        # const  age    smoking  bp        hdl     diab.   parents' stroke
        beta_male   = c(-11.699, 0.1153, 0.4981, 0.0149,  -0.4406, 0.879,  0.2933),
        beta_female = c(-7.966,  0.0633, 0.4163, 0.00893, -0.7636, 1.2383, 0.5470)
    ),
    mechanism = expression({
        subset_male   <- status[alive == 1L & stroke == 0L & sex == 0L, which = TRUE]
        subset_female <- status[alive == 1L & stroke == 0L & sex == 1L, which = TRUE]
        X_male   <- as.matrix(status[subset_male,   
            .(const = 1, age, smoking, blood_pressure, hdl_cholesterol, diabetes, parent_stroke)])
        X_female <- as.matrix(status[subset_female, 
            .(const = 1, age, smoking, blood_pressure, hdl_cholesterol, diabetes, parent_stroke)])

        # Compute 1 day risk from 10 year risk by assuming a Poisson process.
        prob_male_10year   <- 1 - pmin(expit(as.vector(X_male %*% beta_male)), 1)
        prob_female_10year <- 1 - pmin(expit(as.vector(X_female %*% beta_female)), 1)
        lambda_male   <- -log(pmax(prob_male_10year, 1e-12)) / 3650.0
        lambda_female <- -log(pmax(prob_female_10year, 1e-12)) / 3650.0
        new_stroke_male   <- rbern(1 - exp(-lambda_male))
        new_stroke_female <- rbern(1 - exp(-lambda_female))

        # Assign dates for stroke occurrence, update stroke status
        status[subset_male,   `:=` (age_at_stroke = age * new_stroke_male,   stroke = new_stroke_male)]
        status[subset_female, `:=` (age_at_stroke = age * new_stroke_female, stroke = new_stroke_female)]
    })
)

get_diabetes <- ManipulationEvent$new(
    name = "Diabetes",
    description = "Risk of getting diabetes",
    parameters = list(
                # const   age(45-54) age(55-64) bmi(25-30) bmi(>30) w. circ. (1) w.circ. (2) h. glucose
        beta = c(-5.514,  0.628,     0.892,     0.165,     1.096,   0.857,       1.350,      2.139)
    ),
    mechanism = expression({
        subset <- status[alive == 1L & diabetes == 0L & age >= 45, which = TRUE]
        X <- as.matrix(status[subset, .(
            const = 1,
            age1 = (age >= 45 & age < 54), 
            age2 = (age >= 54), 
            bmi1 = (bmi > 25 & bmi <= 30), 
            bmi2 = (bmi > 30), 
            waist_circumference1 = (sex == 0L) * (waist_circumference >= 94 & waist_circumference < 102) +
                                   (sex == 1L) * (waist_circumference >= 80 & waist_circumference < 88),
            waist_circumference2 = (sex == 0L) * (waist_circumference >= 102) + 
                                   (sex == 1L) * (waist_circumference >= 88),
            high_glucose)
        ])

        # Compute 1 day risk from 10 year risk by assuming a Poisson process.
        prob_10year <- 1 - pmin(expit(as.vector(X %*% beta)), 1)
        lambda <- -log(pmax(prob_10year, 1e-12)) / 3650.0
        new_diabetes <- rbern(1 - exp(-lambda))

        # Update diabetes status
        status[subset, diabetes := new_diabetes]
    })
)

stroke_mortality <- ManipulationEvent$new(
    name = "Stroke Mortality",
    description = "Risk of death after suffering a stroke",
    parameters = list(
        initial = 5.41859, 
        longterm = c(8.17473693, -2.47553590, -0.01651868)
    ),
    mechanism = expression({
        u <- status[alive == 1L & stroke == 1L, rbern(fifelse(
            age - age_at_stroke < (28.0 / 365.0), 
            initial, 
            expit(longterm[1] + longterm[2] * exp(((age - age_at_stroke) * 365 - 27) * longterm[3]))))]
        status[alive == 1L & stroke == 1L, `:=` (alive = u, cause_of_death = 2L * (1L - u))]
    })
)

industry_salt <- ManipulationEvent$new(
    name = "Intervention 1",
    description = "Industry reduces salt content in all products",
    parameters = list(
        effect = -0.97
    ),
    mechanism = expression({
        status[alive == 1L, blood_pressure := blood_pressure + effect]
    })
)

doctor_advice <- ManipulationEvent$new(
    name = "Intervention 2",
    description = "Advice to stop adding salt in cooking and at table",
    parameters = list(
        adherence = 0.5,
        effect = -2
    ),
    mechanism = expression({
        status[alive == 1L & blood_pressure >= 140, 
               blood_pressure := blood_pressure + effect * rbern2(.N, adherence)]
    })
)

# Further event organizing
stroke_events  <- list(stroke_mortality, get_stroke)
attr(stroke_events, "ordered") <- TRUE
other_events <- list(stroke_events, other_mortality, get_diabetes)
all_events <- list(aging, other_events)
attr(all_events, "ordered") <- TRUE
