## Bernoulli distribution with probability of success (1) 'prob' (vectorized over prob)
rbern <- function(prob) {
    return(1L * (dqrng::dqrunif(length(prob)) < prob))
}

## Bernoulli distribution for 'n' repetitions with the same probability 'prob'
rbern2 <- function(n, prob) {
    return(1L * (dqrng::dqrunif(n) < prob))
}

## Uniformly random date between 'start' and 'end' dates (inclusive)
rdate <- function(start, end, ...) {
    return(sample(seq(as.Date(start), as.Date(end), by = "day"), ...))
}

## Random binary, where the probability of 1 is 'prob'
rbin <- function(size, prob = 0.5) {
    return(sample(c(0L, 1L), size = size, replace = TRUE, prob = c(1 - prob, prob)))
}

## The logistic link function
logit <- function(x) log(x) - log(1 - x)

## Inverse logistic link
expit <- function(x) (1 + exp(-x))^(-1)

## Function to generate binary starting values based on age-based proportions
## n        : number of values to generate
## sex      : vector of sex for each individual
## age      : vector of ages (in years)
## val_m    : proportions for males in each age group
## val_f    : proportions for females in each age group
## age_min  : youngest age group, no individual should have an age below this
## age_max  : oldest distinct age group, ages above this are considered a single group
## age_step : size of each age group
sex_age_init <- function(n, sex, age, val_m, val_f, age_min, age_max, age_step) {
    groups <- (age_max - age_min) %/% age_step + 1
    age_group <- pmin((age - age_min) %/% age_step + 1, groups)
    age_group[age_group < 1] <- NA
    p <- (age_group > 0) * ((sex == 0) * val_m[age_group] + (sex == 1) * val_f[age_group])
    p[is.na(p)] <- 0
    return(rbinom(n, 1, p))
}

## Function to resample values within an allowed range
## X       : a matrix of values
## sex     : a vector of sex indicators
## smoking : a vector of smoking indicators
## age     : a vector of age groups
## limits  : a list of allowed ranges for column of 'X'
impose_limits <- function(X, sex, smoking, age, limits) {
    age <- as.numeric(age)
    groups <- expand.grid(unique(sex), unique(smoking), unique(age))
    for (i in 1:nrow(groups)) {
        g <- sex == groups[i,1] & smoking == groups[i,2] & age == groups[i,3]
        for (j in 1:ncol(X)) {
            out <- X[g,j] < limits[[j]][1] | X[g,j] > limits[[j]][2]
            bound <- !out
            g_out <- g
            g_out[g] <- out
            g_bound <- g
            g_bound[g] <- !out
            X[g_out,j] <- sample(X[g_bound,j], sum(out), replace = TRUE)
        }
    }
    return(X)
}

## Objective function to calibrate total mortality to mortality of the Finnish population
## x : current vector of parameter values
mortality_objective <- function(x, calib_sim) {
    mortality <- read.csv(file = "data/kuol_007_201700.csv", header = TRUE, sep = ",")
    mortality[,3] <- mortality[,3]/1000
    n <- nrow(mortality)
    # Impute 100-year-olds from 99
    mortality <- rbind(mortality, c(100, 2017, 2 * mortality[n,3] - mortality[n-1,3])) 
    p <- list("Other Mortality" = x[1:2]^2, "Stroke Mortality" = list(x[3]^2, 1))
    calib_res <- calib_sim$configure(t_sim = 364, p = p, output = mortality_df, min_age = 30)
    out <- calib_res$out
    stroke_prop <- calib_res$stroke_prop
    s_obj <- 1e7
    if (stroke_prop > 0) s_obj <- 2000 * (log(stroke_prop) - log(0.07534936))^2
    mortality <- mortality[mortality$Age %in% out$age_group,]
    w <- rep(1, nrow(out))
    w[out$age_group >= 80] <- 100 # Weight for mortalities >= 60, fit should be better here
    valid <- (out[ ,2] != 0 & out[ ,1] != out[,2] & out[ ,1] != 0)
    all_dead <- out[ ,1] == out[ ,2]
    all_alive <- out[ ,1] == 0
    obj <- sum(w[valid] * (log(mortality[valid,3]) - log(out[valid,1]/out[valid,2]))^2)
    obj <- obj + sum(w[all_dead] * log(mortality[all_dead,3])^2)
    obj <- obj + sum(w[all_alive] * log(mortality[all_alive,3])^2)
    return(obj + s_obj)
}

## A function used to calibrate total mortality
## dt        : the status data.table
## min_age   : the youngest age group
mortality_df <- function(dt, min_age = 0) {
    dt[ ,age_group := floor(age)]
    deaths <- dt[alive == 0, .N, by = age_group]
    size <- dt[ , .N, by = age_group]
    groups <- 0:max(dt$age_group)
    ages <- length(groups)
    out <- data.frame(deaths = rep(0, ages), size = rep(0, ages), age_group = groups)
    out[size$age_group + 1,"size"] <- size$N
    if (any(deaths$N > 0)) {
        out[deaths$age_group + 1,"deaths"] <- deaths$N
    }
    out <- out[out$age_group >= min_age, ]
    stroke_prop <- nrow(dt[alive == 0 & stroke == 1 & 
                               (age - age_at_stroke) < (28/365),]) / nrow(dt[alive == 0,])
    return(list(out = out, stroke_prop = stroke_prop))
}
